import tensorflow as tf
from tensorflow import keras

import tensorflow_quantum as tfq
import math
import cirq
import sympy
import numpy as np
import seaborn as sns
import collections
from sklearn.decomposition import PCA
import matplotlib.pyplot as plt
from cirq.contrib.svg import SVGCircuit
plt.style.use('ggplot')
np.random.seed(0)

def kernel(data,cx,cy):
    k_data = []
    for point in data:
        dist = np.sqrt((cx-point[0])**2 + (cy-point[1])**2)
        rotation = get_real_angle(point[0]-cx,point[1]-cy)
        k_data.append([dist,rotation])
    return np.array(k_data)


def qubitization(data):
    rotations = []
    for point in data:
        point_rotations = []
        for data_point in point:
            point_rotations.append(2*np.arcsin(np.sqrt(data_point)))
        rotations.append(point_rotations)
    return np.array(rotations)
#-----------------------------------------------------------------------------------------
def get_real_angle(point0,point1):
    value = np.arctan(point1/point0)
    if value < 0:
        #We know that the angle is between 90-180 or 270-360
        if point1 < 0:
            return abs(np.arctan(point1/point0)) + 3*np.pi/2
        else:
            return abs(np.arctan(point1/point0)) + np.pi/2
    elif value > 0:
        if point1 > 0 :
            return np.arctan(point1/point0) + 0
        else:
            return abs(np.arctan(point1/point0)) + np.pi

(x_train, y_train), (x_test, y_test) = tf.keras.datasets.mnist.load_data()
z=5000
x_train = x_train[:z]
y_train = y_train[:z]
# Rescale the images from [0,255] to the [0.0,1.0] range.
x_train, x_test = x_train[..., np.newaxis]/255.0, x_test[..., np.newaxis]/255.0
x_train = x_train.reshape((z,784))
pca = PCA(n_components=4)
pca.fit(x_train)
x_train = pca.transform(x_train)
for i in range(4):
    if x_train[:,i].min() < 0:
        x_train[:,i] -= x_train[:,i].min()
    x_train[:,i] /= x_train[:,i].max()

ind1 = 5
ind2 = 3
# Ind 1 and ind2 are the numbers you want to binary classification between
plt.hist(x_train[:,0][y_train==ind1])
plt.hist(x_train[:,0][y_train==ind2])

#TENSORFLOW QUANTUM
def convert_to_circuit(data):
    """Encode truncated classical image into quantum datapoint."""
    qubits = cirq.GridQubit.rect(1,2)
    circuit = cirq.Circuit()
    circuit.append(cirq.ry(data[0])(qubits[0]))
    circuit.append(cirq.rz(data[1])(qubits[0]))
    circuit.append(cirq.ry(data[2])(qubits[1]))
    circuit.append(cirq.rz(data[3])(qubits[1]))
    return circuit


#TENSORFLOW QUANTUM
def one_qubit_unitary(bit, symbols):
    """Make a Cirq circuit enacting a rotation of the bloch sphere about the X,
    Y and Z axis, that depends on the values in `symbols`.
    """
    return cirq.Circuit(
        cirq.ry(symbols[0])(bit),
        cirq.rz(symbols[1])(bit))

def two_qubit_unitary(bits, symbols):
    """Make a Cirq circuit enacting a rotation of the bloch sphere about the X,
    Y and Z axis, that depends on the values in `symbols`.
    """
    return cirq.Circuit(
        cirq.ry(symbols[0])(bits[0]),
        cirq.rz(symbols[1])(bits[0]),
        cirq.ry(symbols[0])(bits[1]),
        cirq.rz(symbols[1])(bits[1]))

def ent_qubit_unitary(bits,symbols,swap = False):
    if swap:
        return cirq.Circuit(
            cirq.CNotPowGate(exponent=symbols[0])(bits[1],bits[0]),
            cirq.CZPowGate(exponent=symbols[1])(bits[1],bits[0]))
    else:
        return cirq.Circuit(
            cirq.CNotPowGate(exponent=symbols[0])(bits[0],bits[1]),
            cirq.CZPowGate(exponent=symbols[1])(bits[0],bits[1]))

def create_model_circuit(qubits,get_qasm=False,old_circuit = None):
    """Create sequence of alternating convolution and pooling operators
    which gradually shrink over time."""
    if not old_circuit:
        model_circuit = cirq.Circuit()
    else:
        model_circuit = old_circuit
    if get_qasm:
        symbols= qcnn_model.trainable_variables[0].numpy()
    else:
        symbols = sympy.symbols('qconv0:63')
    # Cirq uses sympy.Symbols to map learnable variables. TensorFlow Quantum
    # scans incoming circuits and replaces these with TensorFlow variables.
    model_circuit += one_qubit_unitary(qubits[0], symbols[0:2])
    model_circuit += one_qubit_unitary(qubits[1], symbols[2:4])
    model_circuit += ent_qubit_unitary(qubits[0:2],symbols[4:6],swap=True)
    model_circuit += one_qubit_unitary(qubits[1],symbols[6:8])
    if get_qasm:
        return model_circuit
    return model_circuit




EPOCHS = 10
# Custom accuracy metric.
@tf.function
def custom_accuracy(y_true, y_pred):
    y_true = tf.squeeze(y_true)
    y_pred = tf.map_fn(lambda x: 1.0 if x >= 0 else -1.0, y_pred)
    return tf.keras.backend.mean(tf.keras.backend.equal(y_true, y_pred))

# Quantum Tensorflow
d1 = x_train[y_train==ind1]
d2 = x_train[y_train==ind2]
comb_data = np.concatenate((d1,d2))
circ_data = [convert_to_circuit(x) for x in qubitization(comb_data)]



#QUantum Deep Neural Net
EPOCHS=25
results = {}
ind1 = 2
ind2 = 9

cluster_state_bits = cirq.GridQubit.rect(1, 2)
readout_operators = cirq.Z(cluster_state_bits[-1])


print(f"Evaluating numbers {ind1} and {ind2}:")
data1_labels = np.zeros(len(x_train[y_train==ind1])) + 1
data2_labels = np.zeros(len(x_train[y_train==ind2])) -1
labels = np.concatenate((data1_labels,data2_labels))
d1 = x_train[y_train==ind1]
d2 = x_train[y_train==ind2]
comb_data = np.concatenate((d1,d2))
circ_data = [convert_to_circuit(x) for x in qubitization(comb_data)]
x_train_tfcirc = tfq.convert_to_tensor(circ_data)
excitation_input = tf.keras.Input(shape=(), dtype=tf.dtypes.string)
quantum_model = tfq.layers.PQC(create_model_circuit(cluster_state_bits),
                               readout_operators)(excitation_input)

qcnn_model = tf.keras.Model(inputs=[excitation_input], outputs=[quantum_model])
model_ftq = qcnn_model.compile(optimizer=tf.keras.optimizers.Adam(),
                   loss=tf.losses.mse,
                   metrics=[custom_accuracy])

quantum_history = qcnn_model.fit(x=x_train_tfcirc,
                         y=labels,
                         batch_size=8,
                         epochs=EPOCHS,
                         verbose=2)


# Generating a classical neural network
model = keras.Sequential([
    keras.layers.Dense(2),
    keras.layers.Dense(1,activation='sigmoid')
])

# Compiling a classical neural network
model.compile(optimizer='adam',
              loss=tf.keras.losses.mse,
              metrics=['accuracy'])


from qiskit import QuantumCircuit, Aer, execute
from qiskit import QuantumRegister, ClassicalRegister
from qiskit import QuantumRegister
from qiskit import QuantumCircuit
from qiskit import Aer, execute
from math import pi,log
from qiskit import *

IBMQ.save_account(
        'IBMQ_TOKEN')

IBMQ.load_account()
provider = IBMQ.get_provider(hub='ibm-q-research', group='RESEARCH_GROUP', project='main')
backend = provider.get_backend('ibmq_valencia')

right = 0
wrong = 0

i = 0
for elm in (circ_data):
    out = create_model_circuit(cirq.GridQubit.rect(1, 2), True,circ_data[i])
    qasm_str = str(out.to_qasm())
    qasm_str =qasm_str+str("creg c[1];\nmeasure q[1] -> c[0];")
    qc = QuantumCircuit.from_qasm_str(qasm_str)
    job = execute(qc, backend, shots = 8000)
    result = job.result()
    prop = result.get_counts()
    p0 = (prop['0']/sum(prop.values()))
    p1 = (prop['1']/sum(prop.values()))
    if p0 > p1:
        temp_label = 1
    else:
        temp_label = -1
    if temp_label == labels[i]:
        right += 1
    else:
        wrong += 1
    i=i+1
print(right/(right+wrong))
