import numpy as np
import random
import math
import matplotlib.pyplot as plt
import matplotlib.mlab as mlab
import seaborn as sns
from qiskit import QuantumRegister, ClassicalRegister
from qiskit import QuantumRegister
from qiskit import QuantumCircuit
from qiskit import Aer, execute
from math import pi
from qiskit import *
from qutip import *
from math import log2
from scipy.stats import norm
np.random.seed(0)

def generate_circular_point(r,cx,cy,noise_factor,num_points):
    X = []
    Y = []
    for _ in range(num_points):
        circle_r = np.random.random() * r
        alpha = 2*math.pi*random.random()
        x = circle_r*math.cos(alpha) + cx
        y = circle_r*math.sin(alpha) + cy
        X.append(x)
        Y.append(y)
    return np.array([X,Y])
def generate_not_circular_point(r,cx,cy,noise_factor,num_points,gap_factor,scaling_fac):
    X = []
    Y = []
    for _ in range(num_points):
        circle_r = (np.random.random() * r * scaling_fac) + r*gap_factor
        alpha = 2*math.pi*random.random()
        x = circle_r*math.cos(alpha) + cx
        y = circle_r*math.sin(alpha) + cy
        X.append(x)
        Y.append(y)
    return np.array([X,Y])

def kernel(data,cx,cy):
    k_data = []
    for point in data:
        dist = np.sqrt((cx-point[0])**2 + (cy-point[1])**2)
        rotation = get_real_angle(point[0]-cx,point[1]-cy)
        k_data.append([dist,rotation])
    return np.array(k_data)
def qubitization(data):
    ry = []
    rz = []
    q_data = []
    for point in data:
        ry = 2*np.arcsin(np.sqrt(point[0]))
        rz = point[1]
        q_data.append([ry,rz])
    return np.array(q_data)

def get_real_angle(point0,point1):
    value = np.arctan(point1/point0)
    print(value)
    if value < 0:
        print("Val less than 0")
        #We know that the angle is between 90-180 or 270-360
        if point1 < 0:
            return abs(np.arctan(point1/point0)) + 3*np.pi/2
        else:
            return abs(np.arctan(point1/point0)) + np.pi/2
    elif value > 0:
        if point1 > 0 :
            print("C")
            return np.arctan(point1/point0) + 0
        else:
            print("D")
            return abs(np.arctan(point1/point0)) + np.pi

get_real_angle(-1,1)

cx = 0.5
plt.style.use("ggplot")
cy = 0.5
r = 0.2
data = generate_circular_point(r,cx,cy,1,100).T
data2 = generate_not_circular_point(r,cx,cy,1,100,1.7, 1.6).T

bad_points = data2[:,0]<0
bad_points += data2[:,0]>1
bad_points += data2[:,1]<0
bad_points += data2[:,1]>1
good_points = [True if x == False else False for x in bad_points]
data2 = data2[good_points]
sns.scatterplot(x=data[:,0],y=data[:,1],label="Class 1")
sns.scatterplot(x=data2[:,0],y=data2[:,1],label="Class 0")
plt.ylim([0,1])
plt.xlim([0,1])
plt.savefig("circles_dataset_classical.eps",format="eps")

#Class 1 data
kernel_data = kernel(data2,cx,cy)
new_max = np.max(kernel_data[:,0])
kernel_data[:,0] = kernel_data[:,0]/new_max
q_kernel_data2 = qubitization(kernel_data)

#Class 0 data
kernel_data = kernel(data,cx,cy)
kernel_data[:,0] = kernel_data[:,0]/new_max
q_kernel_data = qubitization(kernel_data)
temp_max = np.max(kernel_data[:,0])
plt.style.use("default")

points = []
k =3
x,y,z = [],[],[]
g = Bloch()
for point in q_kernel_data:
    vert_angle = point[0]
    hor_angle = point[1]
    x.append(np.sin(vert_angle)*np.cos(hor_angle))
    y.append(np.sin(vert_angle)*np.sin(hor_angle))
    z.append(np.cos(vert_angle))


g.add_points([x,y,z])

x,y,z = [],[],[]
for point in q_kernel_data2:
    vert_angle = point[0]
    hor_angle = point[1]
    x.append(np.sin(vert_angle)*np.cos(hor_angle))
    y.append(np.sin(vert_angle)*np.sin(hor_angle))
    z.append(np.cos(vert_angle))

g.add_points([x,y,z])

q = 1
c = 1
right = 0
wrong = 0

from qiskit import *

IBMQ.save_account(
        'IBM-Q TOKEN',overwrite=True)

IBMQ.load_account()
provider = IBMQ.get_provider(hub='ibm-q-research', group='RESEARCH-GROUP', project='main')
backend = provider.get_backend('ibmq_rome')


for point in q_kernel_data:
    circ = QuantumCircuit(q, c)
    circ.ry(point[0], 0)
    circ.rz(point[1], 0)
    circ.measure(0, 0)
    qasm_str = circ.qasm()
    result = execute(circ, backend).result()
    counts = result.get_counts(circ)
    qc = QuantumCircuit.from_qasm_str(qasm_str)
    job = execute(qc, backend, shots=8000)
    result = job.result()
    prop = result.get_counts()
    try:
        p0 = (prop['0'] / sum(prop.values()))
    except:
        p0 = 0
    try:
        p1 = (prop['1'] / sum(prop.values()))
    except:
        p1= 1
    print(p0 > p1)
    try:
        confidence = counts['1'] / (counts['1'] + counts['0'])
    except:
        if '1' in counts.keys():
            confidence = 1
        else:
            confidence = 0
    print(f"Classification : {round(confidence, 1)} | Class : 0")
    if confidence >= 0.5:
        right += 1
print(circ)
circ.qasm()



